/*
 * 
 *  This file is part of Sourceprojects.org Lycia Copyright 2010 noctarius
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 * 
 */
package org.sourceprojects.lycia.fluent;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;

import org.sourceprojects.lycia.Generator;
import org.sourceprojects.lycia.LyciaParser;
import org.sourceprojects.lycia.LyciaConfigurator;
import org.sourceprojects.lycia.LyciaBuilder.GenerationStrategy;
import org.sourceprojects.lycia.exceptions.LyciaInitializationException;
import org.xml.sax.ContentHandler;

@SuppressWarnings("unchecked")
class ConfigurationBuilderHelper {
	private ConfigurationBuilderHelper() {
	}

	static <B> ConfigurationBuilder<B> createValidatingSchema(
			final boolean validatingSchema) {

		return new AbstractConfigurationBuilder<B>() {

			@Override
			public void configure(LyciaParser<B> xmlParser) {
				xmlParser.setValidatingSchema(validatingSchema);
			}
		};
	}

	static <B> ConfigurationBuilder<B> createGenerationStrategy(
			final GenerationStrategy generationStrategy,
			final Class<? extends Generator> generatorClass) {

		if (generationStrategy == null)
			throw new IllegalArgumentException(
					"GenerationStrategy cannot be null");

		if (generationStrategy == GenerationStrategy.Customized
				&& generatorClass == null)
			throw new IllegalArgumentException(
					"When using GenerationStrategy.Customized Generator class cannot be null");

		return new AbstractConfigurationBuilder<B>() {

			@Override
			public int getPriority() {
				return Integer.MIN_VALUE + 1;
			}

			@Override
			public void configure(LyciaParser<B> xmlParser) {
				LyciaConfigurator<B> configurator = xmlParser
						.getConfigurator();

				configurator.setGenerationStrategy(generationStrategy);
				if (generationStrategy == GenerationStrategy.Customized)
					configurator.getGeneratorFactory().setCustomGeneratorClass(
							(Class<? extends Generator<B>>) generatorClass);
			}
		};
	}

	static <B> ConfigurationBuilder<B> createXmlParserConfigurator(
			final Class<? extends LyciaConfigurator> configuratorClass) {
		if (configuratorClass == null)
			throw new IllegalArgumentException(
					"XmlParserConfigurator class cannot be null");

		return new AbstractConfigurationBuilder<B>() {

			@Override
			public int getPriority() {
				return Integer.MIN_VALUE;
			}

			@Override
			public void configure(LyciaParser<B> xmlParser) {
				try {
					Constructor<? extends LyciaConfigurator<B>> constructor;

					Class<? extends LyciaConfigurator<B>> clazz = (Class<? extends LyciaConfigurator<B>>) configuratorClass;
					constructor = clazz.getConstructor(LyciaParser.class,
							LyciaConfigurator.class);

					LyciaConfigurator<B> configurator = constructor
							.newInstance(xmlParser, xmlParser.getConfigurator());

					xmlParser.setConfigurator(configurator);
				} catch (InvocationTargetException e) {
					throw new LyciaInitializationException(e);
				} catch (IllegalArgumentException e) {
					throw new LyciaInitializationException(e);
				} catch (InstantiationException e) {
					throw new LyciaInitializationException(e);
				} catch (IllegalAccessException e) {
					throw new LyciaInitializationException(e);
				} catch (SecurityException e) {
					throw new LyciaInitializationException(e);
				} catch (NoSuchMethodException e) {
					throw new LyciaInitializationException(e);
				}
			}
		};
	}

	static <B> ConfigurationBuilder<B> createContentHandler(
			final ContentHandler contentHandler) {

		return new AbstractConfigurationBuilder<B>() {

			@Override
			public void configure(LyciaParser<B> xmlParser) {
				xmlParser.getConfigurator().addContentHandler(contentHandler);
			}
		};
	}

	static abstract class AbstractConfigurationBuilder<B> implements
			ConfigurationBuilder<B> {

		public int getPriority() {
			return 0;
		}
	}
}
