/*
 * 
 *  This file is part of Sourceprojects.org Lycia Copyright 2010 noctarius
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 * 
 */
package org.sourceprojects.lycia.fluent;

import java.lang.annotation.Annotation;
import java.lang.reflect.Method;

import javax.xml.XMLConstants;

import org.sourceprojects.lycia.ConstraintValidator;
import org.sourceprojects.lycia.Generator;
import org.sourceprojects.lycia.TagDefinitionParser;
import org.sourceprojects.lycia.LyciaParser;
import org.sourceprojects.lycia.LyciaConfigurator;
import org.sourceprojects.lycia.LyciaBuilder.GenerationStrategy;
import org.xml.sax.ContentHandler;

@SuppressWarnings("unchecked")
public abstract class FluentBuilder<B> {

	public static final <B> FluentBuilder<B> prepare() {
		return new FluentBuilderImpl<B>();
	}

	public static final ValidationBuilder validation(
			Class<? extends ConstraintValidator<? extends Annotation>> validator) {

		return new ValidationBuilder(validator);
	}

	public static final NamespaceBuilder namespace(String prefix, String uri) {
		return new NamespaceBuilder(prefix, uri);
	}

	public static final MethodBuilder method(final Method method,
			final Object parser) {

		return new MethodBuilder() {
			@Override
			public Method getMethod() {
				return method;
			}

			@Override
			public Object getParser() {
				return parser;
			}
		};
	}

	public static final NamespacesBuilder namespaces(
			NamespaceBuilder... namespaceBuilders) {

		return new NamespacesBuilder(namespaceBuilders);
	}

	public static final ValidationsBuilder validations(
			ValidationBuilder... validationBuilders) {

		return new ValidationsBuilder(validationBuilders);
	}

	public static final XPathEventBuilder xpath(String expression,
			MethodBuilder methodBuilder) {

		return new XPathEventBuilder(expression, null, methodBuilder, null);
	}

	public static final RegexEventBuilder regex(String expression,
			MethodBuilder methodBuilder) {

		return new RegexEventBuilder(expression, XMLConstants.NULL_NS_URI,
				methodBuilder, null);
	}

	public static final TagEventBuilder tag(String expression,
			MethodBuilder methodBuilder) {

		return new TagEventBuilder(expression, XMLConstants.NULL_NS_URI,
				methodBuilder, null);
	}

	public static final XPathEventBuilder xpath(String expression,
			MethodBuilder methodBuilder, NamespacesBuilder namespacesBuilder) {

		return new XPathEventBuilder(expression, namespacesBuilder,
				methodBuilder, null);
	}

	public static final RegexEventBuilder regex(String expression,
			String namespaceUri, MethodBuilder methodBuilder) {

		return new RegexEventBuilder(expression, namespaceUri, methodBuilder,
				null);
	}

	public static final TagEventBuilder tag(String expression,
			String namespaceUri, MethodBuilder methodBuilder) {

		return new TagEventBuilder(expression, namespaceUri, methodBuilder,
				null);
	}

	public static final XPathEventBuilder xpath(String expression,
			MethodBuilder methodBuilder, ValidationsBuilder validationsBuilder) {

		return new XPathEventBuilder(expression, null, methodBuilder,
				validationsBuilder);
	}

	public static final RegexEventBuilder regex(String expression,
			MethodBuilder methodBuilder, ValidationsBuilder validationsBuilder) {

		return new RegexEventBuilder(expression, XMLConstants.NULL_NS_URI,
				methodBuilder, validationsBuilder);
	}

	public static final TagEventBuilder tag(String expression,
			MethodBuilder methodBuilder, ValidationsBuilder validationsBuilder) {

		return new TagEventBuilder(expression, XMLConstants.NULL_NS_URI,
				methodBuilder, validationsBuilder);
	}

	public static final XPathEventBuilder xpath(String expression,
			MethodBuilder methodBuilder, NamespacesBuilder namespacesBuilder,
			ValidationsBuilder validationsBuilder) {

		return new XPathEventBuilder(expression, namespacesBuilder,
				methodBuilder, validationsBuilder);
	}

	public static final RegexEventBuilder regex(String expression,
			String namespaceUri, MethodBuilder methodBuilder,
			ValidationsBuilder validationsBuilder) {

		return new RegexEventBuilder(expression, namespaceUri, methodBuilder,
				validationsBuilder);
	}

	public static final TagEventBuilder tag(String expression,
			String namespaceUri, MethodBuilder methodBuilder,
			ValidationsBuilder validationsBuilder) {

		return new TagEventBuilder(expression, namespaceUri, methodBuilder,
				validationsBuilder);
	}

	public static final ConfigurationBuilder validateSchema(
			boolean validatingSchema) {

		return ConfigurationBuilderHelper
				.createValidatingSchema(validatingSchema);
	}

	public static final ConfigurationBuilder generationStrategy(
			GenerationStrategy generationStrategy) {

		return ConfigurationBuilderHelper.createGenerationStrategy(
				generationStrategy, null);
	}

	public static final ConfigurationBuilder generationStrategy(
			GenerationStrategy generationStrategy,
			Class<? extends Generator> generatorClass) {

		return ConfigurationBuilderHelper.createGenerationStrategy(
				generationStrategy, generatorClass);

	}

	public static final ConfigurationBuilder contentHandler(
			ContentHandler contentHandler) {

		return ConfigurationBuilderHelper.createContentHandler(contentHandler);
	}

	public static final ConfigurationBuilder configurator(
			Class<? extends LyciaConfigurator> configuratorClass) {

		return ConfigurationBuilderHelper
				.createXmlParserConfigurator(configuratorClass);
	}

	public abstract LyciaParser<B> build();

	public abstract FluentBuilder<B> parser(
			TagDefinitionParser<B> tagDefinitionParser);

	public abstract FluentBuilder<B> parser(XPathEventBuilder xpathEventBuilder);

	public abstract FluentBuilder<B> parser(RegexEventBuilder regexEventBuilder);

	public abstract FluentBuilder<B> parser(TagEventBuilder tagEventBuilder);

	public abstract FluentBuilder<B> transformer(
			TagDefinitionParser<B> tagDefinitionParser);

	public abstract FluentBuilder<B> transformer(
			XPathEventBuilder xpathEventBuilder);

	public abstract FluentBuilder<B> transformer(
			RegexEventBuilder regexEventBuilder);

	public abstract FluentBuilder<B> transformer(TagEventBuilder tagEventBuilder);

	public abstract FluentBuilder<B> configure(
			ConfigurationBuilder<B> configurationBuilder);

	public abstract FluentBuilder<B> configure(
			ConfigurationBuilder<B>... configurationBuilders);
}
