/*
 * 
 *  This file is part of Sourceprojects.org Lycia Copyright 2010 noctarius
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 * 
 */
package org.sourceprojects.lycia.fluent;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import org.sourceprojects.lycia.TagDefinitionParser;
import org.sourceprojects.lycia.LyciaParser;
import org.sourceprojects.lycia.LyciaConfigurator;
import org.sourceprojects.lycia.LyciaParserFactory;
import org.sourceprojects.lycia.exceptions.LyciaConfigurationException;
import org.sourceprojects.lycia.fluent.XPathEventBuilder.BuildReturn;
import org.sourceprojects.lycia.internal.XPathAwareXmlParserConfigurator;

@SuppressWarnings("unchecked")
class FluentBuilderImpl<B> extends FluentBuilder<B> {
	// Parser lists
	private final List<TagDefinitionParser<B>> tagParsers = new ArrayList<TagDefinitionParser<B>>();
	private final List<RegexEventBuilder<B>> regexEventParsers = new ArrayList<RegexEventBuilder<B>>();
	private final List<XPathEventBuilder<B>> xpathEventParsers = new ArrayList<XPathEventBuilder<B>>();
	private final List<TagEventBuilder<B>> tagEventParsers = new ArrayList<TagEventBuilder<B>>();
	private final List<ConfigurationBuilder<B>> configurationBuilders = new ArrayList<ConfigurationBuilder<B>>();

	@Override
	public LyciaParser<B> build() {
		LyciaParser<B> xmlParser = LyciaParserFactory.newInstance();

		Collections.sort(configurationBuilders,
				new ConfigurationBuilderPriorityComparator());

		for (ConfigurationBuilder<B> configurationBuilder : configurationBuilders) {
			configurationBuilder.configure(xmlParser);
		}

		LyciaConfigurator<B> configurator = xmlParser.getConfigurator();

		for (TagDefinitionParser<B> tagParser : tagParsers)
			configurator.addTagDefinitionParser(tagParser);

		for (TagEventBuilder<B> tagEventParser : tagEventParsers)
			configurator.addTagDefinitionParser(tagEventParser
					.build(configurator));

		for (RegexEventBuilder<B> regexEventParser : regexEventParsers)
			configurator.addTagDefinitionParser(regexEventParser
					.build(configurator));

		XPathAwareXmlParserConfigurator<B> xpathConfigurator = findXPathAwareConfigurator(configurator);
		for (XPathEventBuilder<B> xpathEventParser : xpathEventParsers) {
			if (xpathConfigurator == null)
				throw new LyciaConfigurationException(
						"No XPath aware XmlParserConfigurator was found but is needed for XPath parsing");

			BuildReturn<B> build = xpathEventParser.build(configurator);
			xpathConfigurator.registerXPathTagDefinitionParser(build.delegate,
					build.tag, build.namespaceURI);
		}

		return xmlParser;
	}

	@Override
	public FluentBuilder<B> parser(TagDefinitionParser<B> tagDefinitionParser) {
		tagParsers.add(tagDefinitionParser);
		return this;
	}

	@Override
	public FluentBuilder<B> parser(XPathEventBuilder xpathEventBuilder) {
		xpathEventParsers.add(xpathEventBuilder);
		return this;
	}

	@Override
	public FluentBuilder<B> parser(RegexEventBuilder regexEventBuilder) {
		regexEventParsers.add(regexEventBuilder);
		return this;
	}

	@Override
	public FluentBuilder<B> parser(TagEventBuilder tagEventBuilder) {
		tagEventParsers.add(tagEventBuilder);
		return this;
	}

	@Override
	public FluentBuilder<B> transformer(
			TagDefinitionParser<B> tagDefinitionParser) {

		tagParsers.add(tagDefinitionParser);
		return this;
	}

	@Override
	public FluentBuilder<B> transformer(XPathEventBuilder xpathEventBuilder) {
		xpathEventParsers.add(xpathEventBuilder);
		return this;
	}

	@Override
	public FluentBuilder<B> transformer(RegexEventBuilder regexEventBuilder) {
		regexEventParsers.add(regexEventBuilder);
		return this;
	}

	@Override
	public FluentBuilder<B> transformer(TagEventBuilder tagEventBuilder) {
		tagEventParsers.add(tagEventBuilder);
		return this;
	}

	@Override
	public FluentBuilder<B> configure(
			ConfigurationBuilder<B> configurationBuilder) {
		configurationBuilders.add(configurationBuilder);
		return this;
	}

	@Override
	public FluentBuilder<B> configure(
			ConfigurationBuilder<B>... configurationBuilders) {
		this.configurationBuilders.addAll(Arrays.asList(configurationBuilders));
		return this;
	}

	private class ConfigurationBuilderPriorityComparator implements
			Comparator<ConfigurationBuilder<B>> {

		@Override
		public int compare(ConfigurationBuilder<B> o1,
				ConfigurationBuilder<B> o2) {

			if (o1.getPriority() == o2.getPriority())
				return 0;

			return o1.getPriority() < o2.getPriority() ? -1 : 1;
		}

	}

	private XPathAwareXmlParserConfigurator<B> findXPathAwareConfigurator(
			LyciaConfigurator<B> configurator) {

		if (configurator instanceof XPathAwareXmlParserConfigurator)
			return (XPathAwareXmlParserConfigurator<B>) configurator;

		if (configurator.getParentXmlParserConfigurator() == null)
			return null;

		return findXPathAwareConfigurator(configurator
				.getParentXmlParserConfigurator());
	}
}
