/*
 * 
 *  This file is part of Sourceprojects.org Lycia Copyright 2010 noctarius
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 * 
 */
package org.sourceprojects.lycia.internal;

import org.sourceprojects.lycia.ParserContext;
import org.sourceprojects.lycia.ResourceResolver;
import org.sourceprojects.lycia.ResourceResolverFactory;
import org.sourceprojects.lycia.LyciaParser;
import org.sourceprojects.lycia.LyciaBuilder;
import org.sourceprojects.lycia.LyciaParserFactory;
import org.xml.sax.ErrorHandler;

/**
 * Basic implementation of the {@link LyciaBuilder} interface.
 * 
 * @author noctarius
 * @since 1.9.1
 */
public class LyciaBuilderImpl<B> implements LyciaBuilder<B> {
	private boolean validatingSchema;
	private ErrorHandler errorHandler;
	private ResourceResolver resourceResolver;
	private ParserContext<B> parserContext;
	private GenerationStrategy generationStrategy = GenerationStrategy.ReflectiveGeneration;
	private ClassLoader classLoader;
	private Class<? extends ResourceResolver> resourceResolverClass;
	private Class<? extends ParserContext<B>> parserContextClass;
	private Class<? extends LyciaParser<B>> xmlParserClass;

	@Override
	public LyciaParser<B> build() {
		LyciaParser<B> xmlParser;

		if (resourceResolver == null) {
			if (resourceResolverClass != null) {
				resourceResolver = ResourceResolverFactory.newInstance(
						resourceResolverClass.getCanonicalName(), classLoader);
			} else {
				resourceResolver = ResourceResolverFactory
						.newInstance(classLoader);
			}
		}

		String parserContextClassName;
		if (parserContext != null) {
			if (parserContextClass != null) {
				parserContextClassName = parserContextClass.getCanonicalName();
			} else {
				parserContextClassName = LyciaParserFactory.STANDARD_PARSERCONTEXT_CLASS;
			}
		} else {
			parserContextClassName = LyciaParserFactory.STANDARD_PARSERCONTEXT_CLASS;
		}

		String xmlParserClassName;
		if (xmlParserClass != null) {
			xmlParserClassName = xmlParserClass.getCanonicalName();
		} else {
			xmlParserClassName = LyciaParserFactory.STANDARD_XMLPARSER_CLASS;
		}

		xmlParser = LyciaParserFactory.newInstance(xmlParserClassName,
				parserContextClassName, classLoader, false);
		xmlParser.setErrorHandler(errorHandler);
		xmlParser.setValidatingSchema(validatingSchema);
		xmlParser.setResourceResolver(resourceResolver);
		xmlParser.getConfigurator().setGenerationStrategy(generationStrategy);

		if (parserContext != null) {
			xmlParser.setParserContext(parserContext);
		}

		return xmlParser;
	}

	@Override
	public LyciaBuilder<B> setClassLoader(ClassLoader classLoader) {
		this.classLoader = classLoader;
		return this;
	}

	@Override
	public LyciaBuilder<B> setGenerationStrategy(
			GenerationStrategy generationStrategy) {
		this.generationStrategy = generationStrategy;
		return this;
	}

	@Override
	public LyciaBuilder<B> setErrorHandler(ErrorHandler errorHandler) {
		this.errorHandler = errorHandler;
		return this;
	}

	@Override
	public LyciaBuilder<B> setParserContext(ParserContext<B> parserContext) {
		this.parserContext = parserContext;
		return this;
	}

	@Override
	public LyciaBuilder<B> setParserContextClass(
			Class<? extends ParserContext<B>> parserContextClass) {
		this.parserContextClass = parserContextClass;
		return this;
	}

	@Override
	public LyciaBuilder<B> setResourceResolver(
			ResourceResolver resourceResolver) {
		this.resourceResolver = resourceResolver;
		return this;
	}

	@Override
	public LyciaBuilder<B> setResourceResolverClass(
			Class<? extends ResourceResolver> resourceResolverClass) {
		this.resourceResolverClass = resourceResolverClass;
		return this;
	}

	@Override
	public LyciaBuilder<B> setValidatingSchema(boolean validatingSchema) {
		this.validatingSchema = validatingSchema;
		return this;
	}

	@Override
	public LyciaBuilder<B> setXmlParserClass(
			Class<? extends LyciaParser<B>> xmlParserClass) {
		this.xmlParserClass = xmlParserClass;
		return this;
	}
}
