/*
 * 
 *  This file is part of Sourceprojects.org Lycia Copyright 2010 noctarius
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 * 
 */
package org.sourceprojects.lycia.internal;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.Reader;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;

import javax.xml.XMLConstants;
import javax.xml.namespace.QName;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.validation.Schema;
import javax.xml.validation.SchemaFactory;

import org.apache.log4j.Logger;
import org.sourceprojects.lycia.DefinitionBuilder;
import org.sourceprojects.lycia.ParserContext;
import org.sourceprojects.lycia.ProgressListener;
import org.sourceprojects.lycia.ResourceResolver;
import org.sourceprojects.lycia.TagDefinitionParser;
import org.sourceprojects.lycia.LyciaParser;
import org.sourceprojects.lycia.LyciaConfigurator;
import org.sourceprojects.lycia.annotations.AfterPropertySet;
import org.sourceprojects.lycia.exceptions.LyciaInitializationException;
import org.sourceprojects.lycia.exceptions.LyciaParsingException;
import org.sourceprojects.lycia.internal.bundled.ReaderInputStream;
import org.sourceprojects.lycia.internal.configuration.RankingXmlParserConfigurator;
import org.sourceprojects.lycia.internal.io.LyciaXmlReader;
import org.sourceprojects.lycia.io.LyciaReader;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.xml.sax.ErrorHandler;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

/**
 * Standard implementation of the {@link LyciaParser} interface includes a set of
 * all basic features.
 * 
 * @author noctarius
 * @since 0.0.1
 * 
 * @param <B>
 *            Type of ContextObject
 */
public class LyciaParserImpl<B> implements LyciaParser<B> {
	private static final Logger LOGGER = Logger.getLogger(LyciaParserImpl.class);

	private final List<ProgressListener> listeners = new ArrayList<ProgressListener>();
	private LyciaConfigurator<B> configurator;
	private ResourceResolver resourceResolver;
	private ErrorHandler errorHandler;
	private ParserContext<B> parserContext;
	private boolean validatingSchema = true;
	private Class<? extends LyciaReader<B>> readerClass;
	private LyciaReader<B> reader;
	private long expirationTime = 100;

	public LyciaParserImpl() {
		this(null);
	}

	public LyciaParserImpl(LyciaConfigurator<B> configurator) {
		if (configurator != null) {
			this.configurator = configurator;
		} else {
			this.configurator = new RankingXmlParserConfigurator<B>(this, null);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.sourceprojects.lycia.XmlParser#addTagDefinitionParser(org.sourceprojects
	 * .lycia.TagDefinitionParser)
	 */
	@Override
	public void addTagDefinitionParser(
			TagDefinitionParser<B> tagDefinitionParser) {
		configurator.addTagDefinitionParser(tagDefinitionParser);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.sourceprojects.lycia.XmlParser#addTagDefinitionParsers(org.sourceprojects
	 * .lycia.TagDefinitionParser<B>[])
	 */
	@Override
	public void addTagDefinitionParsers(
			TagDefinitionParser<B>[] tagDefinitionParsers) {
		configurator.addTagDefinitionParsers(tagDefinitionParsers);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @seeorg.sourceprojects.lycia.XmlParser#removeTagDefinitionParser(org.
	 * sourceprojects.lycia.TagDefinitionParser)
	 */
	@Override
	public void removeTagDefinitionParser(
			TagDefinitionParser<B> tagDefinitionParser) {
		configurator.removeTagDefinitionParser(tagDefinitionParser);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @seeorg.sourceprojects.lycia.XmlParser#removeTagDefinitionParsers(org.
	 * sourceprojects.lycia.TagDefinitionParser<B>[])
	 */
	@Override
	public void removeTagDefinitionParsers(
			TagDefinitionParser<B>[] tagDefinitionParsers) {
		configurator.removeTagDefinitionParsers(tagDefinitionParsers);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.sourceprojects.lycia.XmlParser#parse(java.io.InputStream)
	 */
	@Override
	public void parse(InputStream inputStream) throws LyciaParsingException {
		try {
			// Create the per parser instances
			final Document document = createReader(inputStream);
			final DefinitionBuilder definitionBuilder = new DefinitionBuilderImpl(
					document);

			// Start parsing
			for (Node node = document.getFirstChild(); node != null; node = node
					.getNextSibling()) {
				readNode(node, definitionBuilder);
			}
		} catch (final Exception e) {
			throw new LyciaParsingException("Parsing failed", e);
		} finally {
			// Explicitly cleanup actual parserContext for threadded use
			if (parserContext instanceof InternalParserContext<?>) {
				InternalParserContext<?> context = ((InternalParserContext<?>) parserContext);
				Cache<?> patternCache = context.getContextUtils()
						.getPatternCache();
				Cache<?> xpathCache = context.getContextUtils().getXPathCache();

				if (patternCache != null)
					patternCache.shutdown();
				if (xpathCache != null)
					xpathCache.shutdown();

				context.setContextUtils(new ParserContextUtils());
			}
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.sourceprojects.lycia.XmlParser#parse(java.io.Reader)
	 */
	@Override
	public void parse(Reader reader) throws LyciaParsingException {
		parse(reader, 0);
	}

	public void parse(Reader reader, long length)
			throws LyciaParsingException {
		parse(new ReaderInputStream(reader, length));
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.sourceprojects.lycia.XmlParser#parse(java.lang.String)
	 */
	@Override
	public void parse(String xml) throws LyciaParsingException {
		parse(new DataLengthStringReader(xml, xml.length()));
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.sourceprojects.lycia.XmlParser#parse(java.io.File)
	 */
	@Override
	public void parse(File file) throws LyciaParsingException {
		Reader reader = null;
		try {
			reader = new FileReader(file);
			parse(reader, file.length());
		} catch (final FileNotFoundException e) {
			throw new LyciaParsingException("Parsing failed", e);
		} finally {
			if (reader != null) {
				try {
					reader.close();
				} catch (IOException e) {
					e.printStackTrace();
				}
			}
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.sourceprojects.lycia.XmlParser#transform(java.io.InputStream)
	 */
	@Override
	public Document transform(InputStream inputStream)
			throws LyciaParsingException {
		try {
			// Create the per parser instances
			final Document document = createReader(inputStream);

			// Create the output document
			final Document transformedDocument = createDocumentBuilder()
					.newDocument();

			final DefinitionBuilder definitionBuilder = new DefinitionBuilderImpl(
					transformedDocument);

			// Start transforming
			for (Node node = document.getFirstChild(); node != null; node = node
					.getNextSibling()) {
				final Node result = transformNode(node, definitionBuilder);

				if (result != null) {
					transformedDocument.appendChild(result);
				}
			}

			return transformedDocument;
		} catch (final Exception e) {
			throw new LyciaParsingException("Parsing failed", e);
		} finally {
			// Explicitly cleanup actual parserContext for threadded use
			if (parserContext instanceof InternalParserContext<?>)
				((InternalParserContext<?>) parserContext)
						.setContextUtils(new ParserContextUtils());
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.sourceprojects.lycia.XmlParser#transform(java.lang.String)
	 */
	@Override
	public Document transform(String xml) throws LyciaParsingException {
		return transform(new DataLengthStringReader(xml, xml.length()));
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.sourceprojects.lycia.XmlParser#transform(java.io.Reader)
	 */
	@Override
	public Document transform(Reader reader) throws LyciaParsingException {
		return transform(reader, 0);
	}

	public Document transform(Reader reader, long length)
			throws LyciaParsingException {
		return transform(new ReaderInputStream(reader, length));
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.sourceprojects.lycia.XmlParser#transform(java.io.File)
	 */
	@Override
	public Document transform(File file) throws LyciaParsingException {
		Reader reader = null;
		try {
			reader = new FileReader(file);
			return transform(reader, file.length());
		} catch (final FileNotFoundException e) {
			throw new LyciaParsingException("Parsing failed", e);
		} finally {
			if (reader != null) {
				try {
					reader.close();
				} catch (IOException e) {
					e.printStackTrace();
				}
			}
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.sourceprojects.lycia.XmlParser#getResourceResolver()
	 */
	@Override
	public ResourceResolver getResourceResolver() {
		return resourceResolver;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.sourceprojects.lycia.XmlParser#setResourceResolver(org.sourceprojects
	 * .lycia.ResourceResolver)
	 */
	@Override
	public void setResourceResolver(ResourceResolver resourceResolver) {
		this.resourceResolver = resourceResolver;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.sourceprojects.lycia.XmlParser#getErrorHandler()
	 */
	@Override
	public ErrorHandler getErrorHandler() {
		return errorHandler;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.sourceprojects.lycia.XmlParser#setErrorHandler(org.xml.sax.ErrorHandler
	 * )
	 */
	@Override
	public void setErrorHandler(ErrorHandler errorHandler) {
		this.errorHandler = errorHandler;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.sourceprojects.lycia.XmlParser#getParserContext()
	 */
	@Override
	public ParserContext<B> getParserContext() {
		return parserContext;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.sourceprojects.lycia.XmlParser#setParserContext(org.sourceprojects
	 * .lycia.ParserContext)
	 */
	@Override
	public void setParserContext(ParserContext<B> parserContext) {
		this.parserContext = parserContext;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.sourceprojects.lycia.XmlParser#isValidatingSchema()
	 */
	@Override
	public boolean isValidatingSchema() {
		return validatingSchema;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.sourceprojects.lycia.XmlParser#setValidatingSchema(boolean)
	 */
	@Override
	public void setValidatingSchema(boolean validatingSchema) {
		this.validatingSchema = validatingSchema;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.sourceprojects.lycia.XmlParser#addProgressListener(org.sourceprojects
	 * .lycia.ProgressListener)
	 */
	@Override
	public void addProgressListener(ProgressListener listener) {
		listeners.add(listener);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.sourceprojects.lycia.XmlParser#removeProgressListener(org.sourceprojects
	 * .lycia.ProgressListener)
	 */
	@Override
	public void removeProgressListener(ProgressListener listener) {
		listeners.remove(listener);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.sourceprojects.lycia.XmlParser#setConfigurator(org.sourceprojects
	 * .lycia.XmlParserConfigurator)
	 */
	@Override
	public void setConfigurator(LyciaConfigurator<B> configurator) {
		this.configurator = configurator;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.sourceprojects.lycia.XmlParser#getConfigurator()
	 */
	@Override
	public LyciaConfigurator<B> getConfigurator() {
		return configurator;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.sourceprojects.lycia.XmlParser#getReader()
	 */
	@Override
	public Class<? extends LyciaReader<B>> getReader() {
		return readerClass;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.sourceprojects.lycia.XmlParser#setReader(java.lang.Class)
	 */
	@Override
	public void setReader(Class<? extends LyciaReader<B>> readerClass) {
		this.readerClass = readerClass;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.sourceprojects.lycia.XmlParser#purge()
	 */
	@Override
	public void purge() {
		// On finish force caches to expire and cleanup
		if (parserContext instanceof InternalParserContext<?>) {
			ParserContextUtils utils = ((InternalParserContext<?>) parserContext)
					.getContextUtils();

			utils.getPatternCache().shutdown();
			utils.getXPathCache().shutdown();
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.sourceprojects.lycia.XmlParser#getCacheExpirationTime()
	 */
	@Override
	public long getCacheExpirationTime() {
		return expirationTime;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.sourceprojects.lycia.XmlParser#setCacheExpirationTime(long)
	 */
	@Override
	public void setCacheExpirationTime(long expirationTime) {
		this.expirationTime = expirationTime;
	}

	private synchronized Document createReader(InputStream inputStream)
			throws Exception {
		if (reader == null) {
			initializeReader();
		}

		final Document document = reader.read(new InputSource(
				new ProgressAwareReader(inputStream, listeners)));

		return document;
	}

	private DocumentBuilder createDocumentBuilder() throws Exception {
		final DocumentBuilderFactory factory = DocumentBuilderFactory
				.newInstance();

		// setting options for clean parsing
		factory.setIgnoringComments(true);
		factory.setIgnoringElementContentWhitespace(true);
		factory.setNamespaceAware(true);
		if (isValidatingSchema()) {
			try {
				factory.setSchema(createSchema());
			} catch (final UnsupportedOperationException e) {
				LOGGER.warn("Warning: Schema isn't supported");
				// ignore since it isn't supported by all implementations
			}
		}

		return factory.newDocumentBuilder();
	}

	private Schema createSchema() throws SAXException {
		final SchemaFactory factory = SchemaFactory
				.newInstance(XMLConstants.W3C_XML_SCHEMA_NS_URI);
		factory.setResourceResolver(getResourceResolver());
		factory.setErrorHandler(getErrorHandler());
		return factory.newSchema();
	}

	private void initializeReader() throws LyciaInitializationException {
		if (readerClass == null) {
			reader = new LyciaXmlReader<B>();
		} else {
			try {
				reader = readerClass.newInstance();
			} catch (InstantiationException e) {
				throw new LyciaInitializationException(
						"Reader could not be initialized", e);
			} catch (IllegalAccessException e) {
				throw new LyciaInitializationException(
						"Reader could not be initialized", e);
			}
		}

		reader.setXmlParser(this);

		for (Method method : reader.getClass().getMethods()) {
			AfterPropertySet annotation = method
					.getAnnotation(AfterPropertySet.class);

			if (annotation != null) {
				if (!annotation.value()) {
					executeInitialization(method, reader,
							"Reader could not be initialized");
				} else {
					new Thread(createAsyncExecution(method, reader)).start();
				}
			}
		}
	}

	private void readNode(Node node, DefinitionBuilder definitionBuilder)
			throws LyciaParsingException {
		// If node is an element
		if (node.getNodeType() == Node.ELEMENT_NODE) {
			final String namespaceURI = node.getNamespaceURI();
			final String tagname = node.getLocalName();

			// Try to resolve TagDefinitionParser
			final TagDefinitionParser<B> parser = findParserByNamespaceAndTag(
					namespaceURI, tagname, (Element) node);

			if (parser != null) {
				final QName qName = new QName(namespaceURI, tagname);
				parser.parseElement((Element) node, tagname, qName,
						parserContext, definitionBuilder);
			}
		}

		// If node has children, go through all of them
		if (node.hasChildNodes()) {
			for (Node child = node.getFirstChild(); child != null; child = child
					.getNextSibling()) {
				readNode(child, definitionBuilder);
			}
		}
	}

	private Node transformNode(Node node, DefinitionBuilder definitionBuilder)
			throws LyciaParsingException {
		final String namespaceURI = node.getNamespaceURI();
		final String tagname = node.getLocalName();

		// Try to resolve TagDefinitionParser
		TagDefinitionParser<B> parser = null;
		if (namespaceURI != null) {
			parser = findParserByNamespaceAndTag(namespaceURI, tagname,
					(Element) node);
		}

		Node n;
		if (parser != null) {
			final QName qName = new QName(namespaceURI, tagname);
			n = parser.transformElement((Element) node, tagname, qName,
					parserContext, definitionBuilder);
		} else {
			n = definitionBuilder.getDocument().importNode(node, false);
		}

		// If node has children, go through all of them
		if (node.hasChildNodes()) {
			if (n != null) {
				for (Node child = node.getFirstChild(); child != null; child = child
						.getNextSibling()) {

					final Node result = transformNode(child, definitionBuilder);

					if (result != null) {
						n.appendChild(result);
					}
				}
			}
		}

		return n;
	}

	private TagDefinitionParser<B> findParserByNamespaceAndTag(
			String namespaceURI, String tagname, Element element) {
		TagDefinitionParser<B> result = configurator.findTagDefinitionParser(
				tagname, namespaceURI, element);

		if (LOGGER.isDebugEnabled()) {
			LOGGER.debug("Found TagDefinitionParser '" + result + "' for tag '"
					+ namespaceURI + "->" + tagname);
		}

		return result;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return "XmlParserImpl [configurator=" + configurator
				+ ", errorHandler=" + errorHandler + ", listeners=" + listeners
				+ ", parserContext=" + parserContext + ", resourceResolver="
				+ resourceResolver + ", validatingSchema=" + validatingSchema
				+ "]";
	}

	private Runnable createAsyncExecution(final Method method,
			final LyciaReader<B> reader) {
		return new Runnable() {
			@Override
			public void run() {
				executeInitialization(method, reader,
						"Asynchronous initialization failed");
			}
		};
	}

	private void executeInitialization(Method method, LyciaReader<B> reader,
			String message) {
		try {
			method.invoke(reader);
		} catch (IllegalArgumentException e) {
			throw new LyciaInitializationException(message, e);
		} catch (IllegalAccessException e) {
			throw new LyciaInitializationException(message, e);
		} catch (InvocationTargetException e) {
			throw new LyciaInitializationException(message, e);
		}
	}

}
