/*
 * 
 *  This file is part of Sourceprojects.org Lycia Copyright 2010 noctarius
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 * 
 */
package org.sourceprojects.lycia.internal;

import java.lang.ref.Reference;
import java.lang.ref.ReferenceQueue;
import java.lang.ref.SoftReference;
import java.util.Map;
import java.util.Timer;
import java.util.TimerTask;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.locks.ReentrantLock;

import org.apache.log4j.Logger;

/**
 * Implementation of the {@link Cache} interface with {@link SoftReference}d
 * values which can be removed by GarbageCollector if there is no memory left.
 * 
 * @author noctarius
 * @since 1.9.1
 */
public class SoftReferencedCache<T> implements Cache<T> {
	private static final Logger LOGGER = Logger
			.getLogger(SoftReferencedCache.class);

	private final Map<String, SoftReference<T>> cache = new ConcurrentHashMap<String, SoftReference<T>>();
	private final ReentrantLock lock = new ReentrantLock();
	private final ReferenceQueue<T> queue = new ReferenceQueue<T>();
	private final long expireTime;

	private Timer timer = null;

	public SoftReferencedCache() {
		this(100);
	}

	public SoftReferencedCache(long expireTime) {
		this.expireTime = expireTime;
	}

	@Override
	public T get(String name) {
		removeQueuedSoftReferences();
		SoftReference<T> reference = cache.get(name);
		return reference == null ? null : reference.get();
	}

	@Override
	@SuppressWarnings("unchecked")
	public <R extends T> R get(String name, Class<R> type) {
		removeQueuedSoftReferences();
		SoftReference<T> reference = cache.get(name);
		return reference == null ? null : (R) reference.get();
	}

	@Override
	public long getExpire() {
		return expireTime;
	}

	@Override
	public void put(String name, T obj) {
		put(name, obj, expireTime);
	}

	@Override
	public void put(String name, T obj, long expireTime) {
		if (expireTime == -1) {
			expireTime = this.expireTime;
		}

		initializeTimer();
		cache.put(name, new SoftReference<T>(obj, queue));
		timer.schedule(new RemoverTask(name), expireTime);
	}

	private void initializeTimer() {
		lock.lock();
		try {
			if (timer == null) {
				timer = new Timer(true);
			}
		} finally {
			lock.unlock();
		}
	}

	private synchronized void removeQueuedSoftReferences() {
		Reference<? extends T> reference = null;
		// Try to poll next candidate from queue
		while ((reference = queue.poll()) != null) {
			if (LOGGER.isDebugEnabled()) {
				LOGGER.debug("Key was referenced to be removed from cache");
			}

			String keyToRemove = null;
			for (String key : cache.keySet()) {
				SoftReference<T> r = cache.get(key);

				if (r.equals(reference)) {
					// Found key, save to remove
					keyToRemove = key;
					break;
				}
			}

			// Remove if key was found
			if (keyToRemove != null) {
				cache.remove(keyToRemove);
			}
		}
	}

	private class RemoverTask extends TimerTask {
		private final String key;

		RemoverTask(String key) {
			this.key = key;
		}

		@Override
		public void run() {
			if (LOGGER.isDebugEnabled()) {
				LOGGER.debug("Removing " + key + " from cache " + toString());
			}

			cache.remove(key);
		}
	}

	@Override
	public void shutdown() {
		cache.clear();

		if (timer != null) {
			timer.cancel();
			timer.purge();
		}
	}

	@Override
	public String toString() {
		return "SoftReferencedCache [cache=" + cache + ", expireTime="
				+ expireTime + ", lock=" + lock + ", queue=" + queue
				+ ", timer=" + timer + "]";
	}

	@Override
	public void invalidate() {
		cache.clear();
	}

}
