/*
 * 
 *  This file is part of Sourceprojects.org Lycia Copyright 2010 noctarius
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 * 
 */
package org.sourceprojects.lycia.internal;

import org.sourceprojects.lycia.ParserContext;
import org.sourceprojects.lycia.LyciaParser;

/**
 * A basic, generic, threadsafe implementation of the {@link ParserContext}
 * interface. Makes the internal {@link ParserContextUtils} instance available
 * for holding caches used by the actual {@link LyciaParser} instance.
 * 
 * @author noctarius
 * @since 1.9.5
 * 
 * @param <T>
 *            Type of ContextObject
 */
public class ThreadsafeParserContextImpl<T> implements InternalParserContext<T> {
	private ThreadLocal<ThreadLocalHolder> holders = new ThreadLocal<ThreadLocalHolder>() {
		@Override
		protected synchronized ThreadLocalHolder initialValue() {
			return new ThreadLocalHolder();
		}
	};

	@Override
	public T getContextObject() {
		ThreadLocalHolder holder = holders.get();

		if (holder == null) {
			return null;
		}

		return holder.getContextObject();
	}

	@Override
	public void setContextObject(T contextObject) {
		ThreadLocalHolder holder = holders.get();
		holder.setContextObject(contextObject);
	}

	@Override
	public void setParserState(String parserState) {
		ThreadLocalHolder holder = holders.get();
		holder.setParserState(parserState);
	}

	@Override
	public String getParserState() {
		ThreadLocalHolder holder = holders.get();

		if (holder == null) {
			return null;
		}

		return holder.getParserState();
	}

	public void setContextUtils(ParserContextUtils contextUtils) {
		ThreadLocalHolder holder = holders.get();
		holder.setContextUtils(contextUtils);
	}

	public ParserContextUtils getContextUtils() {
		ThreadLocalHolder holder = holders.get();

		if (holder == null) {
			return null;
		}

		return holder.getContextUtils();
	}
	
	public synchronized void remove() {
		holders.remove();
	}

	private class ThreadLocalHolder {
		private ParserContextUtils contextUtils = new ParserContextUtils();
		private T contextObject;
		private String parserState;

		public void setContextObject(T contextObject) {
			this.contextObject = contextObject;
		}

		public T getContextObject() {
			return contextObject;
		}

		public void setParserState(String parserState) {
			this.parserState = parserState;
		}

		public String getParserState() {
			return parserState;
		}

		public void setContextUtils(ParserContextUtils contextUtils) {
			this.contextUtils = contextUtils;
		}

		public ParserContextUtils getContextUtils() {
			return contextUtils;
		}

	}
}
