/*
 * 
 *  This file is part of Sourceprojects.org Lycia Copyright 2010 noctarius
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 * 
 */
package org.sourceprojects.lycia.internal.configuration;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.xml.namespace.QName;
import javax.xml.xpath.XPathExpression;

import org.apache.log4j.Logger;
import org.sourceprojects.lycia.Delegate;
import org.sourceprojects.lycia.Generator;
import org.sourceprojects.lycia.TagDefinitionParser;
import org.sourceprojects.lycia.LyciaParser;
import org.sourceprojects.lycia.LyciaConfigurator;
import org.sourceprojects.lycia.annotations.Namespace;
import org.sourceprojects.lycia.annotations.Tag;
import org.sourceprojects.lycia.annotations.TagParser;
import org.sourceprojects.lycia.annotations.TagTransformer;
import org.sourceprojects.lycia.exceptions.LyciaConfigurationException;
import org.sourceprojects.lycia.internal.Cache;
import org.sourceprojects.lycia.internal.InternalParserContext;
import org.sourceprojects.lycia.internal.XPathAwareXmlParserConfigurator;
import org.sourceprojects.lycia.internal.configuration.TagConfigurationUtils.MethodDefinition;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

/**
 * {@link LyciaConfigurator} implementation for handling annotated
 * {@link TagDefinitionParser} like @{@link TagParser} or @
 * {@link TagTransformer}.
 * 
 * @author noctarius
 * @since 1.9.1
 * 
 * @param <B>
 *            Type of ContextObject
 */
public class AnnotatedXmlParserConfigurator<B> extends
		BasicXmlParserConfigurator<B> implements
		XPathAwareXmlParserConfigurator<B> {
	private static final Logger LOGGER = Logger
			.getLogger(AnnotatedXmlParserConfigurator.class);

	private final Map<XPathExpression, List<TagDefinitionParser<B>>> xpathMappings = new HashMap<XPathExpression, List<TagDefinitionParser<B>>>();

	public AnnotatedXmlParserConfigurator(LyciaParser<B> xmlParser,
			LyciaConfigurator<B> parent) {
		super(xmlParser, parent);
	}

	@Override
	@SuppressWarnings("unchecked")
	public void addTagDefinitionParser(Object definitionParser) {
		// Now search for annotated methods
		MethodDefinition[] definitions = TagConfigurationUtils
				.findAnnotatedMethods(definitionParser.getClass());

		// Basic test if definitionParser could be a legal TagDefinitionParser
		if (definitions.length == 0
				&& !(definitionParser instanceof TagDefinitionParser)) {
			throw new LyciaConfigurationException("Class "
					+ definitionParser.getClass().getCanonicalName()
					+ " is no legal TagDefinitionParser");
		}

		// If methods found add them as DelegateTagDefinitionParsers
		if (definitions.length > 0) {
			addAnnotatedTagDefinitionParser(definitionParser, definitions);
		}
	}

	@Override
	public void addTagDefinitionParsers(Object[] definitionParsers) {
		for (Object definitionParser : definitionParsers) {
			addTagDefinitionParser(definitionParser);
		}
	}

	@Override
	public void addTagDefinitionParsers(
			TagDefinitionParser<B>[] definitionParsers) {
		for (TagDefinitionParser<B> definitionParser : definitionParsers) {
			addTagDefinitionParser(definitionParser);
		}
	}

	@Override
	public void addTagDefinitionParser(TagDefinitionParser<B> definitionParser) {
		// First try to resolve standard methods
		if (definitionParser.getTags() != null
				&& definitionParser.getTags().length > 0) {
			super.addTagDefinitionParser(definitionParser);
		}

		// Now search for annotated methods
		MethodDefinition[] definitions = TagConfigurationUtils
				.findAnnotatedMethods(definitionParser.getClass());

		if (definitions.length > 0) {
			addAnnotatedTagDefinitionParser(definitionParser, definitions);
		}
	}

	@Override
	public List<TagDefinitionParser<B>> findTagDefinitionParsers(
			QName localElement, Element element) {
		List<TagDefinitionParser<B>> parsers = new ArrayList<TagDefinitionParser<B>>();
		try {
			// Start adding all standard TagDefinitionParsers
			parsers.addAll(super
					.findTagDefinitionParsers(localElement, element));
		} catch (LyciaConfigurationException e) {
			if (!isNamespaceInXPathMappings(localElement.getNamespaceURI())) {
				LOGGER.warn("No standard namespace parser for namespaceURI '"
						+ localElement.getNamespaceURI() + "' registered");

				return Collections.emptyList();
			}

			// Use XPath mapped TagDefinitionParsers if there any (in case of
			// exception)
			parsers.addAll(findXPathMapping(element));
			return parsers;
		}

		// Add XPath mapped TagDefinitionParsers
		parsers.addAll(findXPathMapping(element));

		return parsers;
	}

	@Override
	@SuppressWarnings("unchecked")
	public void registerXPathTagDefinitionParser(Delegate<B> delegate, Tag tag,
			String namespaceURI) {
		for (Namespace namespace : tag.xpath().namespaces()) {
			checkNamespaceResolver(namespace.uri());
		}

		XPathExpression compiled = XPathConfigutationUtils.compileExpression(
				tag.xpath().expression(), tag.xpath().namespaces());

		TagDefinitionParser<B> parser = new DelegateTagDefinitionParser<B>(
				delegate, tag, namespaceURI);
		List<TagDefinitionParser<B>> list = asList(parser);
		xpathMappings.put(compiled, list);
	}

	private boolean isNamespaceInXPathMappings(String namespaceURI) {
		Collection<List<TagDefinitionParser<B>>> parserCollections = xpathMappings
				.values();

		// Test for namespaceURI matching
		for (List<TagDefinitionParser<B>> parsers : parserCollections) {
			for (TagDefinitionParser<B> parser : parsers) {
				if (parser.getNamespace() != null
						&& parser.getNamespace().equals(namespaceURI)) {
					return true;
				}
			}
		}

		// Test XPath mapping handler for namespaceURI
		for (List<TagDefinitionParser<B>> parsers : parserCollections) {
			for (TagDefinitionParser<B> parser : parsers) {
				if (XPathConfigutationUtils.isHandleNamespace(parser,
						namespaceURI)) {
					return true;
				}
			}
		}

		return false;
	}

	private List<TagDefinitionParser<B>> findXPathMapping(Element element) {
		InternalParserContext<B> context = getInternalParserContextRepresentation();
		Cache<NodeList> xpathCache = null;

		// Try to retrieve the cache
		if (context != null) {
			xpathCache = context.getContextUtils().getXPathCache();
		}

		// Try to find mapping
		for (XPathExpression expression : xpathMappings.keySet()) {
			if (XPathConfigutationUtils.matches(element, expression,
					xpathCache, getXmlParser())) {
				return xpathMappings.get(expression);
			}
		}

		return Collections.emptyList();
	}

	@SuppressWarnings("unchecked")
	private void addAnnotatedTagDefinitionParser(Object tagDefinitionParser,
			MethodDefinition[] definitions) {

		// Retrieve the generator implementation
		Generator<B> generator = getGeneratorFactory().newInstance(
				getGenerationStrategy());

		// Add for every found method add TagDefinitionParsers created by the
		// generator implementation
		for (MethodDefinition definition : definitions) {
			Delegate<B> delegate = generator.generate(tagDefinitionParser,
					definition);

			if (delegate == null) {
				throw new LyciaConfigurationException(
						"Could not initialize Annotation-Delegate for method "
								+ definition.getMethod().getName()
								+ " on class "
								+ tagDefinitionParser.getClass()
										.getCanonicalName());
			}

			// Add one delegate per Tag
			for (Tag tag : definition.getTags()) {
				String namespaceURI = tag.namespace();
				if ("".equals(namespaceURI)) {
					if (tagDefinitionParser instanceof TagDefinitionParser<?>) {
						namespaceURI = ((TagDefinitionParser<B>) tagDefinitionParser)
								.getNamespace();
					}
				}

				boolean legalXpath = !XPathConfigutationUtils
						.isNullXPathAnnotation(tag);

				if (legalXpath) {
					registerXPathTagDefinitionParser(delegate, tag,
							namespaceURI);
				} else {
					checkNamespaceResolver(namespaceURI);

					addTagDefinitionParserToContainer(
							new DelegateTagDefinitionParser<B>(delegate, tag,
									namespaceURI), namespaceURI);
				}
			}
		}
	}

	private void checkNamespaceResolver(String namespaceURI) {
		if (getXmlParser().getResourceResolver().resolveResource(null,
				namespaceURI, null, null, null) == null) {
			throw new LyciaConfigurationException(new NullPointerException(
					"Namespace " + namespaceURI + " cannot be resolved"));
		}
	}

	private List<TagDefinitionParser<B>> asList(
			TagDefinitionParser<B>... parsers) {
		List<TagDefinitionParser<B>> list = new ArrayList<TagDefinitionParser<B>>();
		for (int i = 0; i < parsers.length; i++) {
			list.add(parsers[i]);
		}

		return Collections.unmodifiableList(list);
	}
}
