/*
 * 
 *  This file is part of Sourceprojects.org Lycia Copyright 2010 noctarius
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 * 
 */
package org.sourceprojects.lycia.internal.configuration;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.regex.PatternSyntaxException;

import javax.xml.XMLConstants;
import javax.xml.namespace.QName;

import org.apache.log4j.Logger;
import org.sourceprojects.lycia.GeneratorFactory;
import org.sourceprojects.lycia.TagDefinitionParser;
import org.sourceprojects.lycia.LyciaParser;
import org.sourceprojects.lycia.LyciaConfigurator;
import org.sourceprojects.lycia.LyciaBuilder.GenerationStrategy;
import org.sourceprojects.lycia.annotations.Tag;
import org.sourceprojects.lycia.exceptions.LyciaConfigurationException;
import org.sourceprojects.lycia.exceptions.LyciaInitializationException;
import org.sourceprojects.lycia.internal.Cache;
import org.sourceprojects.lycia.internal.InternalParserContext;
import org.sourceprojects.lycia.internal.TagDefinitionParserContainer;
import org.sourceprojects.lycia.internal.generator.GeneratorFactoryImpl;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.xml.sax.ContentHandler;

/**
 * Base implementation of {@link LyciaConfigurator}.
 * 
 * @author noctarius
 * @since 1.9.1
 * 
 * @param <B>
 *            Type of ContextObject
 */
public class BasicXmlParserConfigurator<B> implements LyciaConfigurator<B> {
	private static final Logger LOGGER = Logger
			.getLogger(BasicXmlParserConfigurator.class);

	private static final Pattern DEFAULT_PATTERN = Pattern.compile("\\*");

	private final LyciaConfigurator<B> parent;
	private final TagDefinitionParserContainer<B> tagDefinitionParsers = new TagDefinitionParserContainer<B>();
	private final List<ContentHandler> contentHandlers = new ArrayList<ContentHandler>();
	private final LyciaParser<B> xmlParser;

	private GenerationStrategy generationStrategy = GenerationStrategy.ReflectiveGeneration;
	private GeneratorFactory<B> generatorFactory = new GeneratorFactoryImpl<B>();

	public BasicXmlParserConfigurator(LyciaParser<B> xmlParser,
			LyciaConfigurator<B> parent) {
		this.xmlParser = xmlParser;
		this.parent = parent;
	}

	@Override
	public void addTagDefinitionParser(TagDefinitionParser<B> definitionParser) {
		addTagDefinitionParserInternal(definitionParser);
	}

	@Override
	public void addTagDefinitionParsers(
			TagDefinitionParser<B>[] definitionParsers) {
		for (final TagDefinitionParser<B> definitionParser : definitionParsers) {
			addTagDefinitionParserInternal(definitionParser);
		}
	}

	@Override
	public void addTagDefinitionParser(Object definitionParser) {
		addTagDefinitionParserInternal(definitionParser);
	}

	@Override
	public void addTagDefinitionParsers(Object[] definitionParsers) {
		for (final Object definitionParser : definitionParsers) {
			addTagDefinitionParserInternal(definitionParser);
		}
	}

	@Override
	public TagDefinitionParser<B> findTagDefinitionParser(QName localElement,
			Element element) {
		List<TagDefinitionParser<B>> parsers = findTagDefinitionParsers(
				localElement, element);

		if (parsers.size() == 0) {
			return null;
		}

		return parsers.get(0);
	}

	@Override
	public TagDefinitionParser<B> findTagDefinitionParser(String tagName,
			String namespaceURI, Element element) {
		return findTagDefinitionParser(new QName(namespaceURI, tagName),
				element);
	}

	@Override
	public List<TagDefinitionParser<B>> findTagDefinitionParsers(
			QName localElement, Element element) {
		return findTagDefinitionParserInternal(localElement.getNamespaceURI(),
				localElement.getLocalPart(), element);
	}

	@Override
	public List<TagDefinitionParser<B>> findTagDefinitionParsers(
			String tagName, String namespaceURI, Element element) {
		return findTagDefinitionParsers(new QName(namespaceURI, tagName),
				element);
	}

	@Override
	public void removeTagDefinitionParser(
			TagDefinitionParser<B> definitionParser) {
		removeTagDefinitionParserInternal(definitionParser);
	}

	@Override
	public void removeTagDefinitionParsers(
			TagDefinitionParser<B>[] definitionParsers) {
		for (final TagDefinitionParser<B> definitionParser : definitionParsers) {
			removeTagDefinitionParserInternal(definitionParser);
		}
	}

	@Override
	public void removeTagDefinitionParser(Object definitionParser) {
		removeTagDefinitionParserInternal(definitionParser);
	}

	@Override
	public void removeTagDefinitionParsers(Object[] definitionParsers) {
		for (final Object definitionParser : definitionParsers) {
			removeTagDefinitionParserInternal(definitionParser);
		}
	}

	@Override
	public GenerationStrategy getGenerationStrategy() {
		return generationStrategy;
	}

	@Override
	public void setGenerationStrategy(GenerationStrategy generationStrategy) {
		this.generationStrategy = generationStrategy;
	}

	public LyciaParser<B> getXmlParser() {
		return xmlParser;
	}

	@Override
	public LyciaConfigurator<B> getParentXmlParserConfigurator() {
		return parent;
	}

	@Override
	public GeneratorFactory<B> getGeneratorFactory() {
		return generatorFactory;
	}

	@Override
	public void setGeneratorFactory(GeneratorFactory<B> generatorFactory) {
		this.generatorFactory = generatorFactory;
	}

	@Override
	public void addContentHandler(ContentHandler contentHandler) {
		contentHandlers.add(contentHandler);
	}

	@Override
	public void removeContentHandler(ContentHandler contentHandler) {
		contentHandlers.remove(contentHandler);
	}

	@Override
	public List<ContentHandler> getContentHandlers() {
		return contentHandlers;
	}

	protected void addTagDefinitionParserToContainer(
			TagDefinitionParser<B> tagDefinitionParser, String namespaceURI) {

		for (Tag tag : tagDefinitionParser.getTagAnnotations()) {
			if (isAlreadyRegistered(tag, namespaceURI)) {
				throw new LyciaInitializationException(
						namespaceURI
								+ "=>"
								+ tag.value()
								+ " has already a registered parser. Duplicate parsers "
								+ "for same tag-expression is illegal");
			}
		}

		Set<TagDefinitionParser<B>> parsers = tagDefinitionParsers
				.get(namespaceURI);
		if (parsers == null) {
			parsers = new HashSet<TagDefinitionParser<B>>();
		}

		if (LOGGER.isDebugEnabled()) {
			for (Tag tag : tagDefinitionParser.getTagAnnotations()) {
				LOGGER.debug("Registered TagDefinitionParser '"
						+ tagDefinitionParser + "' for tag '" + tag.namespace()
						+ "->" + tag.value());
			}
		}

		parsers.add(tagDefinitionParser);
		tagDefinitionParsers.put(namespaceURI, parsers);
	}

	protected QName[] generateParentQNames(String tag, String namespaceURI) {
		final String[] tags = tag.split("\\\\");

		final List<QName> qNames = new ArrayList<QName>();
		for (final String t : tags) {
			if (t.contains("{")) {
				final String pNamespaceURI = t.substring(t.indexOf("{") + 1, t
						.indexOf("}"));
				final String pTag = t.substring(t.indexOf("}") + 1);

				qNames.add(new QName(pNamespaceURI, pTag));
			} else {
				qNames.add(new QName(namespaceURI, t));
			}
		}

		return qNames.toArray(new QName[qNames.size()]);
	}

	protected InternalParserContext<B> getInternalParserContextRepresentation() {
		if (!(xmlParser.getParserContext() instanceof InternalParserContext<?>)) {
			return null;
		}

		return (InternalParserContext<B>) xmlParser.getParserContext();
	}

	private boolean isAlreadyRegistered(Tag tag, String namespaceURI) {
		Set<TagDefinitionParser<B>> parsers = tagDefinitionParsers
				.get(namespaceURI);

		if (parsers == null) {
			return false;
		}

		for (TagDefinitionParser<B> parser : parsers) {
			for (Tag parserTag : parser.getTagAnnotations()) {
				if (tag.value().equals(parserTag.value())
						&& tag.namespace().equals(parserTag.namespace())) {
					return true;
				}
			}
		}

		return false;
	}

	private List<TagDefinitionParser<B>> findTagDefinitionParserInternal(
			String namespaceURI, String tagname, Element element) {
		List<TagDefinitionParser<B>> result = new ArrayList<TagDefinitionParser<B>>();

		TagDefinitionParser<B> defaultParser = null;

		final Set<TagDefinitionParser<B>> parsers = tagDefinitionParsers
				.get(namespaceURI);

		if (parsers == null) {
			throw new LyciaConfigurationException(
					"No namespace parsers found for namespace " + namespaceURI);
		}

		if (LOGGER.isDebugEnabled()) {
			LOGGER.debug("Try to find TagDefinitionParser for " + namespaceURI
					+ "::" + tagname);
		}

		for (final TagDefinitionParser<B> parser : parsers) {
			for (final Tag tag : parser.getTagAnnotations()) {
				String expression = tag.value();
				String namespace = "".equals(tag.namespace()) ? namespaceURI
						: tag.namespace();

				Pattern pattern = buildPattern(tag);
				if (pattern == null && expression.contains("\\")) {
					final QName[] qNames = generateParentQNames(expression,
							namespace);

					TagDefinitionParser<B> finalParser = matchesElement(
							element, qNames, parser);

					if (finalParser != null) {
						result.add(finalParser);
					}
				} else if (pattern != null) {
					Matcher matcher = pattern.matcher(tagname);
					if (matcher.matches()) {
						result.add(parser);
					}
				} else if (expression.equals(tagname)
						&& namespace.equals(namespaceURI)) {
					result.add(parser);
				} else if (expression.equals("*")
						&& namespace.equals(namespaceURI)) {
					if (defaultParser != null) {
						throw new LyciaConfigurationException(
								"Only one default TagDefinitionParser can be registered for expression "
										+ expression);
					}

					defaultParser = parser;
					result.add(parser);
				}
			}
		}

		return result;
	}

	private TagDefinitionParser<B> matchesElement(Element element,
			QName[] qNames, TagDefinitionParser<B> parser) {

		if (element == null) {
			return null;
		}

		final QName qName = qNames[qNames.length - 1];
		final String localPart = qName.getLocalPart();
		final String namespacePart = qName.getNamespaceURI();

		if (localPart.equals(element.getLocalName())) {

			if ((namespacePart.equals(XMLConstants.NULL_NS_URI) && element
					.getNamespaceURI() == null)
					|| (namespacePart.equals(element.getNamespaceURI()))) {

				if (qNames.length == 1) {
					return parser;
				}

				Node parent = element.getParentNode();
				if (!(parent instanceof Element)) {
					return null;
				}

				final QName[] temp = new QName[qNames.length - 1];
				System.arraycopy(qNames, 0, temp, 0, temp.length);

				return matchesElement((Element) parent, temp, parser);
			}
		}

		return null;
	}

	@SuppressWarnings("unchecked")
	private void addTagDefinitionParserInternal(Object tagDefinitionParser) {
		if (xmlParser.getResourceResolver() == null) {
			throw new NullPointerException("ResourceResolver is null");
		}

		// Assume we have a old fashioned TagDefinitionParser
		if (tagDefinitionParser instanceof TagDefinitionParser<?>) {
			addOldFashionedTagDefinitionParser((TagDefinitionParser<B>) tagDefinitionParser);
			return;
		}

		// Try to handle tagDefinitionParser by parent
		if (parent != null) {
			parent.addTagDefinitionParser(tagDefinitionParser);
		}
	}

	private void addOldFashionedTagDefinitionParser(
			TagDefinitionParser<B> tagDefinitionParser) {
		final String namespaceURI = tagDefinitionParser.getNamespace();

		if (xmlParser.getResourceResolver().resolveResource(null, namespaceURI,
				null, null, null) == null) {
			throw new LyciaConfigurationException(new NullPointerException(
					"Namespace " + namespaceURI + " cannot be resolved"));
		}

		addTagDefinitionParserToContainer(tagDefinitionParser, namespaceURI);
	}

	private Pattern buildPattern(Tag tag) {
		if (!tag.regex()) {
			return null;
		}

		InternalParserContext<B> context = getInternalParserContextRepresentation();
		Cache<Pattern> patternCache = null;
		if (context != null) {
			patternCache = context.getContextUtils().getPatternCache();
		}

		Pattern pattern = null;

		if (patternCache != null) {
			pattern = patternCache.get(String.valueOf(tag.hashCode()));
		}

		if (pattern == null) {
			try {
				pattern = Pattern.compile(tag.value());
			} catch (PatternSyntaxException e) {
				pattern = DEFAULT_PATTERN;
			}

			if (patternCache != null) {
				patternCache.put(String.valueOf(tag.hashCode()), pattern,
						xmlParser.getCacheExpirationTime());
			}
		}

		return pattern == DEFAULT_PATTERN ? null : pattern;
	}

	@SuppressWarnings("unchecked")
	private void removeTagDefinitionParserInternal(Object tagDefinitionParser) {
		String namespaceURI = null;

		if (tagDefinitionParser instanceof TagDefinitionParser<?>) {
			namespaceURI = ((TagDefinitionParser<B>) tagDefinitionParser)
					.getNamespace();
		}

		final Set<TagDefinitionParser<B>> parsers = tagDefinitionParsers
				.get(namespaceURI);

		if (parsers != null) {
			parsers.remove(tagDefinitionParser);
			tagDefinitionParsers.put(namespaceURI, parsers);
		}
	}

}
