/*
 * 
 *  This file is part of Sourceprojects.org Lycia Copyright 2010 noctarius
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 * 
 */
package org.sourceprojects.lycia.internal.configuration;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;

import javax.xml.namespace.QName;

import org.apache.log4j.Logger;
import org.sourceprojects.lycia.DefinitionBuilder;
import org.sourceprojects.lycia.ParserContext;
import org.sourceprojects.lycia.TagDefinitionParser;
import org.sourceprojects.lycia.annotations.Tag;
import org.sourceprojects.lycia.annotations.TagParser;
import org.sourceprojects.lycia.annotations.TagTransformer;
import org.sourceprojects.lycia.exceptions.LyciaConfigurationException;
import org.sourceprojects.lycia.exceptions.LyciaParsingException;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

/**
 * 
 * @author noctarius
 * @since 1.9.1
 */
public class TagConfigurationUtils {
	private static final Logger LOGGER = Logger
			.getLogger(TagConfigurationUtils.class);

	private static final Class<?>[][] LEGAL_PARSER_METHOD_PARAMETERS = new Class[][] {
			new Class[] { Element.class, ParserContext.class,
					DefinitionBuilder.class },
			new Class[] { Element.class, String.class, ParserContext.class,
					DefinitionBuilder.class },
			new Class[] { Element.class, String.class, QName.class,
					ParserContext.class, DefinitionBuilder.class } };

	private static final Class<?>[][] LEGAL_TRANSFORMER_METHOD_PARAMETERS = new Class[][] {
			new Class[] { Element.class, ParserContext.class,
					DefinitionBuilder.class },
			new Class[] { Element.class, String.class, ParserContext.class,
					DefinitionBuilder.class },
			new Class[] { Element.class, String.class, QName.class,
					ParserContext.class, DefinitionBuilder.class } };

	private static Method BASE_TRANSFORMER_METHOD;
	private static Method BASE_PARSER_METHOD;

	private static boolean isLegalTagHandlerMethod(Method method) {
		Class<?>[][] expectedParameters = null;
		if (method.isAnnotationPresent(TagParser.class)) {
			expectedParameters = LEGAL_PARSER_METHOD_PARAMETERS;
		} else if (method.isAnnotationPresent(TagTransformer.class)) {
			if (!isLegalTransformerReturn(method)) {
				return false;
			}

			expectedParameters = LEGAL_TRANSFORMER_METHOD_PARAMETERS;
		}

		if (expectedParameters == null) {
			return false;
		}

		if (method.getExceptionTypes().length == 0) {
			return false;
		}

		boolean found = false;
		for (final Class<?> exception : method.getExceptionTypes()) {
			if (LyciaParsingException.class.isAssignableFrom(exception)) {
				found = true;
			}
		}

		if (!found) {
			return false;
		}

		final Class<?>[] givenParameters = method.getParameterTypes();
		for (final Class<?>[] parameters : expectedParameters) {
			if (isCompatibleSignature(parameters, givenParameters)) {
				return true;
			}
		}

		return false;
	}

	private static boolean isCompatibleSignature(Class<?>[] expected,
			Class<?>[] actual) {
		if (expected.length != actual.length) {
			return false;
		}

		for (int i = 0; i < expected.length; i++) {
			if (!expected[i].isAssignableFrom(actual[i])) {
				return false;
			}
		}

		return true;
	}

	private static boolean isLegalTransformerReturn(Method method) {
		if (method.getReturnType() == Node.class) {
			return true;
		}

		return false;
	}

	public static MethodDefinition[] findAnnotatedMethods(Class<?> clazz) {
		boolean foundDefaultTagTransformer = false;
		boolean foundDefaultTagParser = false;

		final List<MethodDefinition> foundMethods = new ArrayList<MethodDefinition>();
		for (final Method method : clazz.getMethods()) {
			if (method.isAnnotationPresent(TagParser.class)) {
				if (LOGGER.isDebugEnabled()) {
					LOGGER.debug("Found @TagParser annotation on method '"
							+ method + "'");
				}

				final TagParser annotation = method
						.getAnnotation(TagParser.class);
				final Tag[] tags = annotation.value();

				if (hasDefaultTag(tags)) {
					if (foundDefaultTagParser) {
						throw new LyciaConfigurationException(
								"Duplicate default TagParser found on class "
										+ clazz.getCanonicalName());
					} else {
						foundDefaultTagParser = true;
					}
				}

				if (isLegalTagHandlerMethod(method)) {
					if (LOGGER.isDebugEnabled()) {
						LOGGER.debug("Method '" + method
								+ "' qualified as legal TagParser method");
					}

					MethodDefinition definition = new MethodDefinition(method,
							tags);
					foundMethods.add(definition);

				} else {
					throw new LyciaConfigurationException(
							"Illegal method signature found for TagParser "
									+ method.getName() + " on class "
									+ clazz.getCanonicalName());
				}
			} else if (method.isAnnotationPresent(TagTransformer.class)) {
				if (LOGGER.isDebugEnabled()) {
					LOGGER.debug("Found @TagTransformer annotation on method '"
							+ method + "'");
				}

				final TagTransformer annotation = method
						.getAnnotation(TagTransformer.class);
				final Tag[] tags = annotation.value();

				if (hasDefaultTag(tags)) {
					if (foundDefaultTagTransformer) {
						throw new LyciaConfigurationException(
								"Duplicate default TagTransformer found on class "
										+ clazz.getCanonicalName());
					} else {
						foundDefaultTagTransformer = true;
					}
				}

				if (isLegalTagHandlerMethod(method)) {
					if (LOGGER.isDebugEnabled()) {
						LOGGER.debug("Method '" + method
								+ "' qualified as legal TagTransformer method");
					}

					MethodDefinition definition = new MethodDefinition(method,
							tags);
					foundMethods.add(definition);

				} else {
					throw new LyciaConfigurationException(
							"Illegal method signature found for TagTransformer "
									+ method.getName() + " on class "
									+ clazz.getCanonicalName());
				}
			}

		}

		return foundMethods.toArray(new MethodDefinition[foundMethods.size()]);
	}

	private static boolean hasDefaultTag(Tag[] tags) {
		for (final Tag tag : tags) {
			if ("*".equals(tag.value())
					&& XPathConfigutationUtils.isNullXPathAnnotation(tag)) {
				return true;
			}
		}

		return false;
	}

	public static synchronized Method getBaseTransformerMethod()
			throws SecurityException, NoSuchMethodException {
		if (BASE_TRANSFORMER_METHOD == null) {
			BASE_TRANSFORMER_METHOD = TagDefinitionParser.class.getMethod(
					"transformElement", LEGAL_TRANSFORMER_METHOD_PARAMETERS[2]);
		}

		return BASE_TRANSFORMER_METHOD;
	}

	public static synchronized Method getBaseParserMethod()
			throws SecurityException, NoSuchMethodException {
		if (BASE_PARSER_METHOD == null) {
			BASE_PARSER_METHOD = TagDefinitionParser.class.getMethod(
					"parseElement", LEGAL_PARSER_METHOD_PARAMETERS[2]);
		}

		return BASE_PARSER_METHOD;
	}

	public static <B> Object[] generateParameter(Method method,
			Element element, String localName, QName parent,
			ParserContext<B> parserContext, DefinitionBuilder definitionBuilder) {
		Object[] result = new Object[method.getParameterTypes().length];

		if (LOGGER.isDebugEnabled()) {
			LOGGER.debug("Resolving parameters for method '" + method + "'");
		}

		switch (method.getParameterTypes().length) {
		case 3:
			result[0] = element;
			result[1] = parserContext;
			result[2] = definitionBuilder;
			break;

		case 4:
			result[0] = element;
			result[1] = localName;
			result[2] = parserContext;
			result[3] = definitionBuilder;
			break;

		case 5:
			result[0] = element;
			result[1] = localName;
			result[2] = parent;
			result[3] = parserContext;
			result[4] = definitionBuilder;
			break;
		default:
		}

		return result;
	}

	public static class MethodDefinition {
		private final Tag[] tags;
		private final Method method;

		public MethodDefinition(Method method, Tag... tags) {
			this.method = method;
			this.tags = tags;
		}

		public Tag[] getTags() {
			Tag[] result = new Tag[tags.length];
			System.arraycopy(tags, 0, result, 0, tags.length);
			return result;
		}

		public Method getMethod() {
			return method;
		}

	}
}
