/*
 * 
 *  This file is part of Sourceprojects.org Lycia Copyright 2010 noctarius
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 * 
 */
package org.sourceprojects.lycia.internal.configuration;

import java.beans.IntrospectionException;
import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import javax.xml.namespace.QName;

import org.sourceprojects.lycia.TagDefinitionParser;
import org.sourceprojects.lycia.LyciaParser;
import org.sourceprojects.lycia.LyciaConfigurator;
import org.sourceprojects.lycia.annotations.Tag;
import org.sourceprojects.lycia.annotations.Validator;
import org.sourceprojects.lycia.exceptions.LyciaConfigurationException;
import org.sourceprojects.lycia.internal.validation.ConstraintValidatorUtils;
import org.sourceprojects.lycia.internal.validation.PropertyDescriptorImpl;
import org.w3c.dom.Element;

/**
 * 
 * @author noctarius
 * @since 1.9.1
 * 
 * @param <B>
 */
public class ValidationAwareXmlParserConfigurator<B> extends
		AnnotatedXmlParserConfigurator<B> {

	public ValidationAwareXmlParserConfigurator(LyciaParser<B> xmlParser,
			LyciaConfigurator<B> parent) {
		super(xmlParser, parent);
	}

	@Override
	public List<TagDefinitionParser<B>> findTagDefinitionParsers(
			QName localElement, Element element) {
		List<TagDefinitionParser<B>> parsers = super.findTagDefinitionParsers(
				localElement, element);

		Iterator<TagDefinitionParser<B>> iterator = parsers.iterator();
		while (iterator.hasNext()) {
			TagDefinitionParser<B> parser = iterator.next();
			Tag[] tags = parser.getTagAnnotations();
			for (Tag tag : tags) {
				PropertyDescriptorImpl[] properties = findValidators(tag);
				if (!ConstraintValidatorUtils.checkConstraints(properties,
						element)) {
					iterator.remove();
				}
			}
		}

		return parsers;
	}

	private PropertyDescriptorImpl[] findValidators(Tag tag) {
		if (tag.validations() == null || tag.validations().length == 0) {
			return new PropertyDescriptorImpl[0];
		}

		for (Validator validator : tag.validations()) {
			Class<?> impl = validator.value();
			try {
				return findValidationFields(impl);
			} catch (IntrospectionException e) {
				throw new LyciaConfigurationException(
						"Failure while validating fields", e);
			}
		}

		return new PropertyDescriptorImpl[0];
	}

	private PropertyDescriptorImpl[] findValidationFields(Class<?> validator)
			throws IntrospectionException {
		List<PropertyDescriptorImpl> result = new ArrayList<PropertyDescriptorImpl>();

		for (Field field : validator.getDeclaredFields()) {
			if (ConstraintValidatorUtils.getCompliantAnnotations(field).length > 0) {
				result.add(new PropertyDescriptorImpl(field, validator));
			}
		}

		return result.toArray(new PropertyDescriptorImpl[result.size()]);
	}
}
