/*
 * 
 *  This file is part of Sourceprojects.org Lycia Copyright 2010 noctarius
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 * 
 */
package org.sourceprojects.lycia.internal.generator;

import java.lang.reflect.Method;
import java.util.concurrent.atomic.AtomicLong;

import org.objectweb.asm.ClassVisitor;
import org.objectweb.asm.ClassWriter;
import org.objectweb.asm.MethodVisitor;
import org.objectweb.asm.Opcodes;
import org.objectweb.asm.util.CheckMethodAdapter;
import org.objectweb.asm.util.TraceMethodVisitor;
import org.sourceprojects.lycia.Delegate;
import org.sourceprojects.lycia.Generator;
import org.sourceprojects.lycia.exceptions.LyciaConfigurationException;
import org.sourceprojects.lycia.internal.configuration.TagConfigurationUtils.MethodDefinition;
import org.sourceprojects.lycia.internal.privileged.CreateClassLoader;
import org.w3c.dom.Element;

/**
 * 
 * @author noctarius
 * @since 1.9.1
 * 
 * @param <T>
 */
public class BytecodeGenerator<T> implements Generator<T>, Opcodes {
	private static final String PARSE_SIGNATURE_THREE_PARAMS = "(Lorg/w3c/dom/Element;Lorg/sourceprojects/lycia/ParserContext;Lorg/sourceprojects/lycia/DefinitionBuilder;)V";
	private static final String PARSE_SIGNATURE_FOUR_PARAMS = "(Lorg/w3c/dom/Element;Ljava/lang/String;Lorg/sourceprojects/lycia/ParserContext;Lorg/sourceprojects/lycia/DefinitionBuilder;)V";
	private static final String PARSE_SIGNATURE_FIVE_PARAMS = "(Lorg/w3c/dom/Element;Ljava/lang/String;Ljavax/xml/namespace/QName;Lorg/sourceprojects/lycia/ParserContext;Lorg/sourceprojects/lycia/DefinitionBuilder;)V";
	private static final String TRANSFORM_SIGNATURE_THREE_PARAMS = "(Lorg/w3c/dom/Element;Lorg/sourceprojects/lycia/ParserContext;Lorg/sourceprojects/lycia/DefinitionBuilder;))Lorg/w3c/dom/Node;";
	private static final String TRANSFORM_SIGNATURE_FOUR_PARAMS = "(Lorg/w3c/dom/Element;Ljava/lang/String;Lorg/sourceprojects/lycia/ParserContext;Lorg/sourceprojects/lycia/DefinitionBuilder;)Lorg/w3c/dom/Node;";
	private static final String TRANSFORM_SIGNATURE_FIVE_PARAMS = "(Lorg/w3c/dom/Element;Ljava/lang/String;Ljavax/xml/namespace/QName;Lorg/sourceprojects/lycia/ParserContext;Lorg/sourceprojects/lycia/DefinitionBuilder;)Lorg/w3c/dom/Node;";

	private static final String PARSE_SIGNATURE_DELEGATE = "(Lorg/w3c/dom/Element;Ljava/lang/String;Ljavax/xml/namespace/QName;Lorg/sourceprojects/lycia/ParserContext;Lorg/sourceprojects/lycia/DefinitionBuilder;)V";
	private static final String PARSE_SIGNATURE_DELEGATE_GENERIFIED = "(Lorg/w3c/dom/Element;Ljava/lang/String;Ljavax/xml/namespace/QName;Lorg/sourceprojects/lycia/ParserContext<TT;>;Lorg/sourceprojects/lycia/DefinitionBuilder;)V";
	private static final String TRANSFORM_SIGNATURE_DELEGATE = "(Lorg/w3c/dom/Element;Ljava/lang/String;Ljavax/xml/namespace/QName;Lorg/sourceprojects/lycia/ParserContext;Lorg/sourceprojects/lycia/DefinitionBuilder;)Lorg/w3c/dom/Node;";
	private static final String TRANSFORM_SIGNATURE_DELEGATE_GENERIFIED = "(Lorg/w3c/dom/Element;Ljava/lang/String;Ljavax/xml/namespace/QName;Lorg/sourceprojects/lycia/ParserContext<Ljava/lang/StringBuilder;>;Lorg/sourceprojects/lycia/DefinitionBuilder;)Lorg/w3c/dom/Node;";

	private static final String EXCEPTION_CLASS = "org/sourceprojects/lycia/exceptions/XmlParserParsingException";
	private static final String SUPER_CLASS = "org/sourceprojects/lycia/internal/generator/BytecodeDelegate";
	private static final String PARSE_METHOD_NAME = "invokeParse";
	private static final String TRANSFORM_METHOD_NAME = "invokeTransform";
	private static final int METHOD_ACCESS = ACC_PUBLIC + ACC_FINAL;

	private static final AtomicLong LONG = new AtomicLong(0);

	private final GeneratorClassLoader classloader;

	public BytecodeGenerator() {
		classloader = CreateClassLoader.createClassLoader(getClass()
				.getClassLoader());
	}

	@Override
	public Delegate<T> generate(Object parser, MethodDefinition methodDefinition) {
		try {
			long id = LONG.getAndIncrement();
			StringBuilder classname = new StringBuilder();
			classname.append("BytecodeGenerated").append(id);

			ClassWriter cw = new ClassWriter(0);
			cw.visit(V1_6, ACC_PUBLIC + ACC_SUPER, classname.toString(),
					toGenerifiedInternalRepresentation(classname.toString()),
					SUPER_CLASS, null);

			// Adding standard constructor
			createConstructor(cw);

			// Adding methods
			if (methodDefinition.getMethod().getReturnType() != Element.class) {
				createParseMethod(cw, methodDefinition);
				callAReturn(cw);
			} else {
				callReturn(cw);
				createTransformMethod(cw, methodDefinition);
			}

			// Closing class visit
			cw.visitEnd();

			byte[] bytecode = cw.toByteArray();

			Class<? extends BytecodeDelegate<T>> clazz = classloader
					.loadClass(bytecode);
			BytecodeDelegate<T> instance = clazz.newInstance();
			instance.setParser(parser);
			return instance;
		} catch (IllegalAccessException e) {
			throw new LyciaConfigurationException(
					"BytecodeGenerator could not generate Delegate for parser '"
							+ parser + "'", e);
		} catch (InstantiationException e) {
			throw new LyciaConfigurationException(
					"BytecodeGenerator could not generate Delegate for parser '"
							+ parser + "'", e);
		}
	}

	private MethodVisitor createConstructor(ClassVisitor cw) {
		MethodVisitor mv = cw.visitMethod(ACC_PUBLIC, "<init>", "()V", null,
				null);
		mv.visitCode();
		mv.visitVarInsn(ALOAD, 0);
		mv.visitMethodInsn(INVOKESPECIAL, SUPER_CLASS, "<init>", "()V");
		mv.visitInsn(RETURN);
		mv.visitMaxs(1, 1);
		mv.visitEnd();
		return mv;
	}

	private MethodVisitor createParseMethod(ClassVisitor cw,
			MethodDefinition methodDefinition) {
		MethodVisitor mv = cw.visitMethod(METHOD_ACCESS, PARSE_METHOD_NAME,
				PARSE_SIGNATURE_DELEGATE, PARSE_SIGNATURE_DELEGATE_GENERIFIED,
				new String[] { EXCEPTION_CLASS });
		TraceMethodVisitor tmv = new TraceMethodVisitor(mv);
		CheckMethodAdapter cma = new CheckMethodAdapter(tmv);

		cma.visitCode();
		callParseParameter(cma, methodDefinition.getMethod());
		cma.visitEnd();
		return mv;
	}

	private void callParseParameter(MethodVisitor mv, Method method) {
		String classname = method.getDeclaringClass().getCanonicalName();
		String methodname = method.getName();
		mv.visitVarInsn(ALOAD, 0);
		mv.visitMethodInsn(INVOKEVIRTUAL, SUPER_CLASS, "getParser",
				"()Ljava/lang/Object;");
		mv.visitTypeInsn(CHECKCAST, toInternalRepresentation(classname));
		mv.visitVarInsn(ASTORE, 6);
		mv.visitVarInsn(ALOAD, 6);
		switch (method.getParameterTypes().length) {
		case 3:
			mv.visitVarInsn(ALOAD, 1);
			mv.visitVarInsn(ALOAD, 4);
			mv.visitVarInsn(ALOAD, 5);
			mv.visitMethodInsn(INVOKEVIRTUAL,
					toInternalRepresentation(classname), methodname,
					PARSE_SIGNATURE_THREE_PARAMS);
			break;
		case 4:
			mv.visitVarInsn(ALOAD, 1);
			mv.visitVarInsn(ALOAD, 3);
			mv.visitVarInsn(ALOAD, 4);
			mv.visitVarInsn(ALOAD, 5);
			mv.visitMethodInsn(INVOKEVIRTUAL,
					toInternalRepresentation(classname), methodname,
					PARSE_SIGNATURE_FOUR_PARAMS);
			break;
		default:
			mv.visitVarInsn(ALOAD, 1);
			mv.visitVarInsn(ALOAD, 2);
			mv.visitVarInsn(ALOAD, 3);
			mv.visitVarInsn(ALOAD, 4);
			mv.visitVarInsn(ALOAD, 5);
			mv.visitMethodInsn(INVOKEVIRTUAL,
					toInternalRepresentation(classname), methodname,
					PARSE_SIGNATURE_FIVE_PARAMS);
		}
		mv.visitInsn(RETURN);
		mv.visitMaxs(8, 7);
	}

	private MethodVisitor createTransformMethod(ClassVisitor cw,
			MethodDefinition methodDefinition) {
		MethodVisitor mv = cw.visitMethod(METHOD_ACCESS, TRANSFORM_METHOD_NAME,
				TRANSFORM_SIGNATURE_DELEGATE,
				TRANSFORM_SIGNATURE_DELEGATE_GENERIFIED,
				new String[] { EXCEPTION_CLASS });

		mv.visitCode();
		callTransformParameter(mv, methodDefinition.getMethod());
		mv.visitEnd();
		return mv;
	}

	private void callTransformParameter(MethodVisitor mv, Method method) {
		String classname = method.getDeclaringClass().getCanonicalName();
		String methodname = method.getName();
		mv.visitVarInsn(ALOAD, 0);
		mv.visitMethodInsn(INVOKEVIRTUAL, SUPER_CLASS, "getParser",
				"()Ljava/lang/Object;");
		mv.visitTypeInsn(CHECKCAST, toInternalRepresentation(classname));
		mv.visitVarInsn(ASTORE, 6);
		mv.visitVarInsn(ALOAD, 6);
		switch (method.getParameterTypes().length) {
		case 3:
			mv.visitVarInsn(ALOAD, 1);
			mv.visitVarInsn(ALOAD, 4);
			mv.visitVarInsn(ALOAD, 5);
			mv.visitMethodInsn(INVOKEVIRTUAL,
					toInternalRepresentation(classname), methodname,
					TRANSFORM_SIGNATURE_THREE_PARAMS);
			break;
		case 4:
			mv.visitVarInsn(ALOAD, 1);
			mv.visitVarInsn(ALOAD, 3);
			mv.visitVarInsn(ALOAD, 4);
			mv.visitVarInsn(ALOAD, 5);
			mv.visitMethodInsn(INVOKEVIRTUAL,
					toInternalRepresentation(classname), methodname,
					TRANSFORM_SIGNATURE_FOUR_PARAMS);
			break;
		default:
			mv.visitVarInsn(ALOAD, 1);
			mv.visitVarInsn(ALOAD, 2);
			mv.visitVarInsn(ALOAD, 3);
			mv.visitVarInsn(ALOAD, 4);
			mv.visitVarInsn(ALOAD, 5);
			mv.visitMethodInsn(INVOKEVIRTUAL,
					toInternalRepresentation(classname), methodname,
					TRANSFORM_SIGNATURE_FIVE_PARAMS);
		}
		mv.visitInsn(RETURN);
		mv.visitMaxs(4, 7);
	}

	private String toInternalRepresentation(String classname) {
		return classname.replace(".", "/");
	}

	private String toGenerifiedInternalRepresentation(String classname) {
		return "<T:Ljava/lang/Object;>L" + toInternalRepresentation(classname)
				+ "<TT;>;";
	}

	private void callReturn(ClassVisitor cw) {
		MethodVisitor mv = cw.visitMethod(METHOD_ACCESS, PARSE_METHOD_NAME,
				PARSE_SIGNATURE_DELEGATE, null,
				new String[] { EXCEPTION_CLASS });
		mv.visitInsn(RETURN);
		mv.visitMaxs(0, 6);
		mv.visitEnd();
	}

	private void callAReturn(ClassVisitor cw) {
		MethodVisitor mv = cw.visitMethod(METHOD_ACCESS, TRANSFORM_METHOD_NAME,
				TRANSFORM_SIGNATURE_DELEGATE, null,
				new String[] { EXCEPTION_CLASS });
		mv.visitCode();
		mv.visitInsn(ACONST_NULL);
		mv.visitInsn(ARETURN);
		mv.visitMaxs(1, 6);
		mv.visitEnd();
	}
}
